@@ -231,8 +231,10 @@ gnutls_certificate_type_set_priority(gnutls_session_t session,
 }
 
 static const int supported_ecc_normal[] = {
+#ifdef ENABLE_NON_SUITEB_CURVES
 	GNUTLS_ECC_CURVE_SECP192R1,
 	GNUTLS_ECC_CURVE_SECP224R1,
+#endif
 	GNUTLS_ECC_CURVE_SECP256R1,
 	GNUTLS_ECC_CURVE_SECP384R1,
 	GNUTLS_ECC_CURVE_SECP521R1,
@@ -1077,6 +1079,7 @@ gnutls_priority_init(gnutls_priority_t * priority_cache,
 	}
 	gnutls_free(darg);
 	gnutls_free(*priority_cache);
+	*priority_cache = NULL;
 
 	return GNUTLS_E_INVALID_REQUEST;
 
