@@ -34,6 +34,7 @@
 #include <gnutls_datum.h>
 #include <x509_int.h>
 #include <common.h>
+#include <pkcs11_int.h>
 #include <gnutls_pk.h>
 
 /* Checks if two certs are identical.  Return 1 on match. */
@@ -75,6 +76,31 @@ _gnutls_check_if_same_cert(gnutls_x509_crt_t cert1,
 	return result;
 }
 
+int
+_gnutls_check_if_same_cert2(gnutls_x509_crt_t cert1,
+			    gnutls_datum_t * cert2bin)
+{
+	gnutls_datum_t cert1bin = { NULL, 0 };
+	int result;
+
+	result = _gnutls_x509_der_encode(cert1->cert, "", &cert1bin, 0);
+	if (result < 0) {
+		result = 0;
+		gnutls_assert();
+		goto cleanup;
+	}
+
+	if ((cert1bin.size == cert2bin->size) &&
+	    (memcmp(cert1bin.data, cert2bin->data, cert1bin.size) == 0))
+		result = 1;
+	else
+		result = 0;
+
+      cleanup:
+	_gnutls_free_datum(&cert1bin);
+	return result;
+}
+
 /* Checks if the issuer of a certificate is a
  * Certificate Authority, or if the certificate is the same
  * as the issuer (and therefore it doesn't need to be a CA).
@@ -682,6 +708,118 @@ _gnutls_x509_verify_certificate(const gnutls_x509_crt_t * certificate_list,
 	return 0;
 }
 
+#ifdef ENABLE_PKCS11
+/* Verify X.509 certificate chain using a PKCS #11 token.
+ *
+ * Note that the return value is an OR of GNUTLS_CERT_* elements.
+ *
+ * This function verifies a X.509 certificate list. The certificate
+ * list should lead to a trusted certificate in order to be trusted.
+ */
+unsigned int
+_gnutls_pkcs11_verify_certificate(const char* url,
+				const gnutls_x509_crt_t * certificate_list,
+				unsigned clist_size,
+				unsigned int flags,
+				gnutls_verify_output_function func)
+{
+	int ret;
+	unsigned int status = 0, i;
+	gnutls_x509_crt_t issuer = NULL;
+	gnutls_datum_t raw_issuer = {NULL, 0};
+
+	if (clist_size > 1) {
+		/* Check if the last certificate in the path is self signed.
+		 * In that case ignore it (a certificate is trusted only if it
+		 * leads to a trusted party by us, not the server's).
+		 *
+		 * This prevents from verifying self signed certificates against
+		 * themselves. This (although not bad) caused verification
+		 * failures on some root self signed certificates that use the
+		 * MD2 algorithm.
+		 */
+		if (gnutls_x509_crt_check_issuer
+		    (certificate_list[clist_size - 1],
+		     certificate_list[clist_size - 1]) != 0) {
+			clist_size--;
+		}
+	}
+
+	/* We want to shorten the chain by removing the cert that matches
+	 * one of the certs we trust and all the certs after that i.e. if
+	 * cert chain is A signed-by B signed-by C signed-by D (signed-by
+	 * self-signed E but already removed above), and we trust B, remove
+	 * B, C and D. */
+	if (!(flags & GNUTLS_VERIFY_DO_NOT_ALLOW_SAME))
+		i = 0;		/* also replace the first one */
+	else
+		i = 1;		/* do not replace the first one */
+
+	for (; i < clist_size; i++) {
+		if (_gnutls_pkcs11_crt_is_known (url, certificate_list[i], 
+			GNUTLS_PKCS11_OBJ_FLAG_COMPARE|GNUTLS_PKCS11_OBJ_FLAG_RETRIEVE_TRUSTED) != 0) {
+			clist_size = i;
+			break;
+		}
+		/* clist_size may have been changed which gets out of loop */
+	}
+
+	if (clist_size == 0) {
+		/* The certificate is already present in the trusted certificate list.
+		 * Nothing to verify. */
+		return status;
+	}
+
+	/* check for blacklists */
+	for (i = 0; i < clist_size; i++) {
+		if (_gnutls_pkcs11_crt_is_known (url, certificate_list[i], 
+			GNUTLS_PKCS11_OBJ_FLAG_RETRIEVE_DISTRUSTED) != 0) {
+			status |= GNUTLS_CERT_INVALID;
+			status |= GNUTLS_CERT_REVOKED;
+			if (func)
+				func(certificate_list[i], certificate_list[i], NULL, status);
+			goto cleanup;
+		}
+	}
+
+	/* check against issuer */
+	ret = gnutls_pkcs11_get_raw_issuer(url, certificate_list[clist_size - 1],
+		&raw_issuer, GNUTLS_X509_FMT_DER, 0);
+	if (ret < 0) {
+		gnutls_assert();
+		status |= GNUTLS_CERT_INVALID;
+		status |= GNUTLS_CERT_SIGNER_NOT_FOUND;
+		return status;
+	}
+
+	ret = gnutls_x509_crt_init(&issuer);
+	if (ret < 0) {
+		gnutls_assert();
+		status |= GNUTLS_CERT_INVALID;
+		status |= GNUTLS_CERT_SIGNER_NOT_FOUND;
+		goto cleanup;
+	}
+
+	ret = gnutls_x509_crt_import(issuer, &raw_issuer, GNUTLS_X509_FMT_DER);
+	if (ret < 0) {
+		gnutls_assert();
+		status |= GNUTLS_CERT_INVALID;
+		status |= GNUTLS_CERT_SIGNER_NOT_FOUND;
+		goto cleanup;
+	}
+	
+	status = _gnutls_x509_verify_certificate(certificate_list, clist_size,
+				&issuer, 1, flags, func);
+
+cleanup:
+	gnutls_free(raw_issuer.data);
+	if (issuer != NULL)
+		gnutls_x509_crt_deinit(issuer);
+	
+	return status;
+}
+#endif
+
 /* This will return the appropriate hash to verify the given signature.
  * If signature is NULL it will return an (or the) appropriate hash for
  * the given parameters.
